/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstamaenc.h"
#include "gstamacommon.h"
#include "gstxmaframeallocator.h"
#include "gstxrm.h"

#include <stdio.h>
#include "ama_dynamic_params.h"

#include <xmaerror.h>
#include <xmasidedata.h>

GST_DEBUG_CATEGORY_STATIC (gst_ama_enc_debug);
#define GST_CAT_DEFAULT gst_ama_enc_debug

#define GST_AMA_ENC_PARMA_COUNT 20
#define GST_AMA_ENC_BITRATE_DEFAULT   0
#define GST_AMA_ENC_MIN_QP_DEFAULT  XMA_ENC_MIN_QP_DEFAULT
#define GST_AMA_ENC_GOP_LENGTH_DEFAULT  XMA_ENC_GOP_SIZE_DEFAULT
#define GST_AMA_ENC_IDR_INTERVAL_DEFAULT  G_MAXUINT32
#define GST_AMA_ENC_LOOKAHEAD_DEPTH_MAX   XMA_ENC_MAX_LOOKAHEAD_DEPTH
#define GST_AMA_ENC_LOOKAHEAD_DEPTH_DEFAULT   -1
#define GST_AMA_ENC_SLICE_DEFAULT   -1
#define GST_AMA_ENC_MAX_BITRATE_DEFAULT   -1
#define GST_AMA_ENC_B_FRAMES_DEFAULT  XMA_ENC_BF_DEFAULT
#define GST_AMA_ENC_FORCED_IDR_DEFAULT  XMA_ENC_FORCED_IDR_DEFAULT
#define GST_AMA_ENC_QP_DEFAULT  -1
#define GST_AMA_ENC_SPATIAL_AQ_DEFAULT  XMA_ENC_SPATIAL_AQ_DEFAULT
#define GST_AMA_ENC_SPATIAL_AQ_GAIN_DEFAULT   XMA_ENC_SPATIAL_AQ_GAIN_DEFAULT
#define GST_AMA_ENC_TEMPORAL_AQ_DEFAULT   XMA_ENC_TEMPORAL_AQ_DEFAULT
#define GST_AMA_ENC_TEMPORAL_AQ_GAIN_DEFAULT  XMA_ENC_TEMPORAL_AQ_GAIN_DEFAULT
#define GST_AMA_ENC_LATENCY_LOGGING_DEFAULT   XMA_ENC_LATENCY_LOGGING_DEFAULT
#define GST_AMA_ENC_CRF_DEFAULT   XMA_ENC_PARAM_CRF_DEFAULT
#define GST_AMA_ENC_DYNAMIC_GOP_DEFAULT   XMA_ENC_DYNAMIC_GOP_DEFAULT
#define GST_AMA_ENCODING_PARAMS_FILE_DEFAULT  NULL
#define GST_AMA_ENC_LATENCY_MS_DEFAULT  XMA_ENC_LATENCY_MS_DEFAULT
#define GST_AMA_ENC_BUFSIZE_DEFAULT   XMA_ENC_BUFSIZE_DEFAULT
#define GST_AMA_ENC_MIN_BITRATE_DEFAULT   -1
#define GST_AMA_ENC_NO_LOWLAT_BFRAMES_DEFAULT XMA_ENC_NO_LOWLAT_BFRAMES_DEFAULT

typedef enum {
  GST_AMA_ENC_PRESET_SUPER_FAST = XMA_ENC_PRESET_SUPER_FAST,
  GST_AMA_ENC_PRESET_FAST       = XMA_ENC_PRESET_FAST,
  GST_AMA_ENC_PRESET_MEDIUM     = XMA_ENC_PRESET_MEDIUM,
  GST_AMA_ENC_PRESET_SLOW       = XMA_ENC_PRESET_SLOW,
  GST_AMA_ENC_PRESET_SUPER_SLOW = XMA_ENC_PRESET_SUPER_SLOW,
  GST_AMA_ENC_PRESET_DEFAULT    = XMA_ENC_PRESET_DEFAULT
} GstAMAEncPreset;

#define GST_TYPE_AMA_ENC_PRESET (gst_ama_enc_preset_get_type ())
static GType
gst_ama_enc_preset_get_type (void)
{
  static GType type = 0;

  static const GEnumValue presets[] = {
    /*{GST_AMA_ENC_PRESET_SUPER_FAST, "Super fast", "super_fast"},*/
    {GST_AMA_ENC_PRESET_FAST, "Fast", "fast"},
    {GST_AMA_ENC_PRESET_MEDIUM, "Medium", "medium"},
    {GST_AMA_ENC_PRESET_SLOW, "Slow", "slow"},
    /*{GST_AMA_ENC_PRESET_SUPER_SLOW, "Super slow", "super_slow"},*/
    {0, NULL, NULL},
  };

  if (g_once_init_enter (&type)) {
    GType tmp = g_enum_register_static ("GstAMAEncPreset", presets);
    g_once_init_leave (&type, tmp);
  }

  return type;
}

typedef enum {
  GST_AMA_ENC_RC_AUTO = XMA_ENC_RC_MODE_AUTO,
  GST_AMA_ENC_RC_CONSTANT_QP = XMA_ENC_RC_MODE_CONSTANT_QP,
  GST_AMA_ENC_RC_CBR = XMA_ENC_RC_MODE_CBR,
  GST_AMA_ENC_RC_VBR = XMA_ENC_RC_MODE_VBR,
  GST_AMA_ENC_RC_CVBR = XMA_ENC_RC_MODE_CVBR,
  GST_AMA_ENC_RC_CABR = XMA_ENC_RC_MODE_CABR,
  GST_AMA_ENC_RC_DEFAULT = XMA_ENC_RC_MODE_DEFAULT
} GstAMAEncRateControl;

#define GST_TYPE_AMA_ENC_RATE_CONTROL (gst_ama_enc_rate_control_get_type ())
static GType
gst_ama_enc_rate_control_get_type (void)
{
  static GType type = 0;

  static const GEnumValue ratecontrols[] = {
    {GST_AMA_ENC_RC_AUTO, "Auto", "auto"},
    {GST_AMA_ENC_RC_CONSTANT_QP, "Constant QP", "cqp"},
    {GST_AMA_ENC_RC_CBR, "Constant bitrate", "cbr"},
    {GST_AMA_ENC_RC_VBR, "Variable bitrate", "vbr"},
    {GST_AMA_ENC_RC_CVBR, "Constrained variable bitrate", "cvbr"},
    {GST_AMA_ENC_RC_CABR, "Content adaptive bitrate", "cabr"},
    {0, NULL, NULL},
  };

  if (g_once_init_enter (&type)) {
    GType tmp = g_enum_register_static ("GstAMAEncRateControl", ratecontrols);
    g_once_init_leave (&type, tmp);
  }

  return type;
}

typedef enum
{
  GST_AMA_ENC_QP_MODE_AUTO = XMA_ENC_QP_MODE_AUTO,
  GST_AMA_ENC_QP_MODE_RELATIVE_LOAD = XMA_ENC_QP_MODE_RELATIVE_LOAD,
  GST_AMA_ENC_QP_MODE_UNIFORM = XMA_ENC_QP_MODE_UNIFORM,
  GST_AMA_ENC_QP_MODE_DEFAULT = XMA_ENC_QP_MODE_DEFAULT,
} GstAMAEncQPMode;


#define GST_TYPE_AMA_ENC_QP_MODE (gst_ama_enc_qp_mode_get_type ())
static GType
gst_ama_enc_qp_mode_get_type (void)
{
  static GType type = 0;

  static const GEnumValue modes[] = {
    {GST_AMA_ENC_QP_MODE_AUTO, "Auto", "auto"},
    {GST_AMA_ENC_QP_MODE_RELATIVE_LOAD, "Relative load (spatial AQ + temporal AQ)", "relative-load"},
    {GST_AMA_ENC_QP_MODE_UNIFORM, "Uniform (AQ off)", "uniform"},
    {0, NULL, NULL},
  };

  if (g_once_init_enter (&type)) {
    GType tmp = g_enum_register_static ("GstAMAEncQPMode", modes);
    g_once_init_leave (&type, tmp);
  }

  return type;
}

typedef enum
{
  GST_AMA_ENC_TUNE_METRICS_VQ = XMA_ENC_TUNE_METRICS_VQ,
  GST_AMA_ENC_TUNE_METRICS_PSNR = XMA_ENC_TUNE_METRICS_PSNR,
  GST_AMA_ENC_TUNE_METRICS_SSIM = XMA_ENC_TUNE_METRICS_SSIM,
  GST_AMA_ENC_TUNE_METRICS_VMAF = XMA_ENC_TUNE_METRICS_VMAF,
  GST_AMA_ENC_TUNE_METRICS_DEFAULT = XMA_ENC_TUNE_METRICS_DEFAULT
} GstAMAEncTuneMetrics;

#define GST_TYPE_AMA_ENC_TUNE_METRICS (gst_ama_enc_tune_metrics_get_type ())
static GType
gst_ama_enc_tune_metrics_get_type (void)
{
  static GType type = 0;

  static const GEnumValue metrics[] = {
    {GST_AMA_ENC_TUNE_METRICS_VQ, "Best visual quality", "vq"},
    {GST_AMA_ENC_TUNE_METRICS_PSNR, "Best PSNR", "psnr"},
    {GST_AMA_ENC_TUNE_METRICS_SSIM, "Best SSIM", "ssim"},
    {GST_AMA_ENC_TUNE_METRICS_VMAF, "Best VMAF", "vmaf"},
    {0, NULL, NULL},
  };

  if (g_once_init_enter (&type)) {
    GType tmp = g_enum_register_static ("GstAMAEncTuneMetrics", metrics);
    g_once_init_leave (&type, tmp);
  }

  return type;
}

typedef enum
{
  GST_AMA_ENC_TIER_AUTO = -1,
  GST_AMA_ENC_TIER_MAIN = XMA_ENC_TIER_MAIN,
  GST_AMA_ENC_TIER_HIGH = XMA_ENC_TIER_HIGH,
  GST_AMA_ENC_TIER_DEFAULT = XMA_ENC_TIER_DEFAULT
} GstAMAEncTier;

#define GST_TYPE_AMA_ENC_TIER (gst_ama_enc_tier_get_type ())
static GType
gst_ama_enc_tier_get_type (void)
{
  static GType type = 0;

  static const GEnumValue tiers[] = {
    {GST_AMA_ENC_TIER_AUTO, "Auto", "auto"},
    {GST_AMA_ENC_TIER_MAIN, "Main", "main"},
    {GST_AMA_ENC_TIER_HIGH, "High", "high"},
    {0, NULL, NULL},
  };

  if (g_once_init_enter (&type)) {
    GType tmp = g_enum_register_static ("GstAMAEncTier", tiers);
    g_once_init_leave (&type, tmp);
  }

  return type;
}

enum
{
  PROP_0,
  PROP_DEVICE,
  PROP_PRESET,
  PROP_BITRATE,
  PROP_GOP_LENGTH,
  PROP_IDR_INTERVAL,
  PROP_LOOKAHEAD_DEPTH,
  PROP_SLICE,
  PROP_RATE_CONTROL,
  PROP_MAX_BITRATE,
  PROP_FORCED_IDR,
  PROP_QP_MODE,
  PROP_SPATIAL_AQ,
  PROP_SPATIAL_AQ_GAIN,
  PROP_TEMPORAL_AQ,
  PROP_TEMPORAL_AQ_GAIN,
  PROP_TUNE_METRICS,
  PROP_TIER,
  PROP_LATENCY_LOGGING,
  PROP_CRF,
  PROP_DYNAMIC_GOP,
  PROP_ENCODING_PARAMS_FILE,
  PROP_LATENCY_MS,
  PROP_BUFSIZE,
  PROP_MIN_BITRATE,

  /* Properties with different ranges in subclasses. */
  PROP_B_FRAMES = GST_AMA_ENC_RANGED_PROPERTY_FIRST,
  PROP_QP,
  PROP_MIN_QP,
  PROP_MAX_QP,
  PROP_NO_LOWLAT_BFRAMES,
  PROP_QP_I_OFFSET,
  PROP_QP_B_OFFSET,
};

enum
{
  SIGNAL_0,
  FRAME_RECEIVED_SIGNAL,
  LAST_SIGNAL,
};

static guint signals[LAST_SIGNAL] = { 0 };

static GstStaticPadTemplate sink_template = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS
    ("video/x-raw(" GST_CAPS_FEATURE_MEMORY_AMA_MEMORY "), "
        "format = (string) { NV12_4L4, I420, NV12_10LE32, P010_10LE, I420_10LE, NV12_10LE_4L4 }, "
        "width = (int) [ " G_STRINGIFY (GST_AMA_MIN_HEIGHT) ", 7680 ], " \
        "height = (int) [ " G_STRINGIFY (GST_AMA_MIN_HEIGHT) ", 7680 ], " \
        "framerate = (fraction) [ 0/1, MAX ];")
    );

typedef struct
{
  GstXma *xma;
  GstXrm *xrm;
  XmaEncoderSession *encoder_session;

  int32_t preset;
  guint32 bitrate;
  int64_t max_bitrate;
  gint32 qp;
  gint32 min_qp;
  gint32 max_qp;
  gint32 gop_length;
  guint32 idr_interval;
  gint32 lookahead_depth;
  int32_t slice;
  int32_t b_frames;
  int32_t forced_idr;
  int32_t qp_mode;
  int32_t spatial_aq;
  int32_t spatial_aq_gain;
  int32_t temporal_aq;
  int32_t temporal_aq_gain;
  int32_t tune_metrics;
  int32_t tier;
  int32_t latency_logging;
  int32_t crf;
  GstAMAEncRateControl rate_control;
  int32_t dynamic_gop;
  gchar *encoding_params_file;
  int32_t latency_ms;
  int32_t bufsize;
  FILE *encoding_params_stream;
  int32_t no_low_latency_b_frames;
  int64_t min_bitrate;
  gint32 qp_i_offset;
  gint32 qp_b_offset;
} GstAMAEncPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (GstAMAEnc, gst_ama_enc, GST_TYPE_VIDEO_ENCODER);

static gboolean
gst_ama_enc_start (GstVideoEncoder * encoder)
{
  GstAMAEnc *self = GST_AMA_ENC (encoder);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  if (priv->encoding_params_file) {
    priv->encoding_params_stream = fopen (priv->encoding_params_file, "r");
    if (!priv->encoding_params_stream) {
      GST_ERROR_OBJECT (self,
          "Unable to open encoding parameters file '%s' for reading: %d",
          priv->encoding_params_file, errno);
      return FALSE;
    }
  }

  return TRUE;
}

static gboolean
gst_ama_enc_close_session (GstAMAEnc * self)
{
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  if (priv->encoder_session) {
    if (xma_enc_session_destroy (priv->encoder_session) != XMA_SUCCESS) {
      GST_ELEMENT_ERROR (self, LIBRARY, SHUTDOWN,
          ("Failed to destroy encoder session"), (NULL));
      return FALSE;
    }
    priv->encoder_session = NULL;
  }

  g_clear_pointer (&priv->xrm, gst_xrm_free);

  return TRUE;
}

static gboolean
gst_ama_enc_stop (GstVideoEncoder * encoder)
{
  GstAMAEnc *self = GST_AMA_ENC (encoder);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  if (!gst_ama_enc_close_session (self)) {
    return FALSE;
  }

  g_clear_pointer (&priv->encoding_params_stream, fclose);
  g_clear_object (&priv->xma);

  return TRUE;
}

static gboolean
gst_ama_enc_set_format (GstVideoEncoder * encoder,
    GstVideoCodecState * state)
{
  GstAMAEnc *self = GST_AMA_ENC (encoder);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);
  GstAMAEncClass *klass = GST_AMA_ENC_GET_CLASS (self);

  gint device;
  gint width, height;
  GstCaps *allowed_caps;
  GstCaps *outcaps;
  GstVideoCodecState *output_state;
  XmaEncoderProperties enc_props = {0};
  XmaParameter params[GST_AMA_ENC_PARMA_COUNT];

  if (!gst_ama_enc_close_session (self)) {
    return FALSE;
  }

  device = gst_ama_device_get_from_caps (state->caps);
  if (device == GST_XMA_DEVICE_UNSPECIFIED) {
    g_clear_object (&priv->xma);
    GST_ELEMENT_ERROR (self, CORE, CAPS,
        ("Invalid or missing 'ama-device' in sink caps"), (NULL));
    return FALSE;
  }

  if (!priv->xma || priv->xma->device != device) {
    g_clear_object (&priv->xma);

    priv->xma = gst_xma_get_for_device (device);
    if (!priv->xma) {
      GST_ELEMENT_ERROR (self, LIBRARY, INIT,
          ("Failed to initialize XMA library"), (NULL));
      return FALSE;
    }
  }

  width = GST_VIDEO_INFO_WIDTH (&state->info);
  height = GST_VIDEO_INFO_HEIGHT (&state->info);

  if (MIN (width, height) > 4320) {
    GST_ELEMENT_ERROR (self, CORE, NEGOTIATION,
        ("Maximum allowed resolution is 7680x4320 or 4320x7680"), (NULL));
    return FALSE;
  }

  if (MAX (width, height) < GST_AMA_MIN_WIDTH) {
      GST_ELEMENT_ERROR (self, CORE, NEGOTIATION,
          ("Minimum allowed resolution is "
          G_STRINGIFY (GST_AMA_MIN_WIDTH) "x" G_STRINGIFY (GST_AMA_MIN_HEIGHT)
          " or "
          G_STRINGIFY (GST_AMA_MIN_HEIGHT) "x" G_STRINGIFY (GST_AMA_MIN_WIDTH)),
          (NULL));
      return FALSE;
    }

  allowed_caps = gst_pad_get_allowed_caps (GST_VIDEO_ENCODER_SRC_PAD (encoder));

  GST_DEBUG_OBJECT (self, "Allowed caps: %" GST_PTR_FORMAT, allowed_caps);

  if (!allowed_caps) {
    return FALSE;
  }

  outcaps = gst_caps_copy (allowed_caps);

  enc_props.profile        = -1; /* auto profile */
  enc_props.param_cnt      = 0;
  enc_props.params         = params;
  /*
  For XAV1
  gst_ama_av1_enc_configure_encoder: Updates 2 additional parameters,
  from klass->configure_encoder
  */

  klass->configure_encoder (self, outcaps, &enc_props);

  gst_clear_caps (&allowed_caps);

  {
    GstStructure *s = gst_caps_get_structure (outcaps, 0);

    if (enc_props.level == 0) {
      gst_structure_remove_field (s, "level");
    }
    if (enc_props.profile == -1) {
      gst_structure_remove_field (s, "profile");
    }
  }

  gst_caps_set_simple (outcaps,
      "bit-depth-luma", G_TYPE_UINT, GST_VIDEO_INFO_COMP_DEPTH (&state->info, 0),
      "bit-depth-chroma", G_TYPE_UINT, GST_VIDEO_INFO_COMP_DEPTH (&state->info, 0),
      NULL);

  GST_DEBUG_OBJECT (self, "Setting output caps: %" GST_PTR_FORMAT, outcaps);

  output_state = gst_video_encoder_set_output_state (GST_VIDEO_ENCODER (self),
      outcaps, state);
  gst_video_codec_state_unref (output_state);
  params[enc_props.param_cnt].name = XMA_ENC_PARAM_SLICE;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->slice;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_MAX_BITRATE;
  params[enc_props.param_cnt].type = XMA_INT64;
  params[enc_props.param_cnt].length = sizeof (int64_t);
  params[enc_props.param_cnt].value = &priv->max_bitrate;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_BF;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->b_frames;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_QP_MODE;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->qp_mode;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_FORCED_IDR;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->forced_idr;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_SPATIAL_AQ;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->spatial_aq;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_TEMPORAL_AQ;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->temporal_aq;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_TUNE_METRICS;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->tune_metrics;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_TIER;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->tier;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_LATENCY_LOGGING;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->latency_logging;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_CRF;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->crf;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_DYNAMIC_GOP;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->dynamic_gop;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_LATENCY_MS;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->latency_ms;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_BUFSIZE;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->bufsize;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_NO_LOWLAT_BFRAMES;
  params[enc_props.param_cnt].type = XMA_INT32;
  params[enc_props.param_cnt].length = sizeof (int32_t);
  params[enc_props.param_cnt].value = &priv->no_low_latency_b_frames;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_MIN_BITRATE;
  params[enc_props.param_cnt].type = XMA_INT64;
  params[enc_props.param_cnt].length = sizeof (int64_t);
  params[enc_props.param_cnt].value = &priv->min_bitrate;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_QP_I_OFFSET;
  params[enc_props.param_cnt].type = XMA_INT64;
  params[enc_props.param_cnt].length = sizeof (gint32);
  params[enc_props.param_cnt].value = &priv->qp_i_offset;
  enc_props.param_cnt++;

  params[enc_props.param_cnt].name = XMA_ENC_PARAM_QP_B_OFFSET;
  params[enc_props.param_cnt].type = XMA_INT64;
  params[enc_props.param_cnt].length = sizeof (gint32);
  params[enc_props.param_cnt].value = &priv->qp_b_offset;
  enc_props.param_cnt++;

  g_assert (enc_props.param_cnt <= G_N_ELEMENTS (params));

  enc_props.handle = priv->xma->xma;
  enc_props.format = XMA_VPE_FMT_TYPE;
  enc_props.sw_format = gst_xma_get_format (GST_VIDEO_INFO_FORMAT (&state->info),
      (gint32 *) &enc_props.flags);
  enc_props.width = width;
  enc_props.height = height;
  enc_props.framerate.numerator = GST_VIDEO_INFO_FPS_N (&state->info);
  enc_props.framerate.denominator = GST_VIDEO_INFO_FPS_D (&state->info);
  enc_props.preset = priv->preset;
  enc_props.bitrate = priv->bitrate;
  enc_props.qp = priv->qp;
  enc_props.gop_size = priv->gop_length;
  enc_props.idr_interval = priv->idr_interval;
  enc_props.lookahead_depth = priv->lookahead_depth;
  enc_props.minQP = priv->min_qp;
  enc_props.maxQP = priv->max_qp;
  enc_props.rc_mode = priv->rate_control;
  enc_props.spat_aq_gain = priv->spatial_aq_gain;
  enc_props.temp_aq_gain = priv->temporal_aq_gain;

  priv->xrm = gst_xrm_reserve_enc_resources (priv->xma->device, &priv->slice,
      &enc_props);

  if (!priv->xrm) {
    GST_ELEMENT_ERROR (self, RESOURCE, FAILED,
        ("Unable to reserve XRM resources for encoding %dx%d@%d/%d",
        enc_props.width, enc_props.height, enc_props.framerate.numerator,
        enc_props.framerate.denominator), (NULL));
    return FALSE;
  }

  priv->encoder_session = xma_enc_session_create (&enc_props);
  if (!priv->encoder_session) {
    GST_ELEMENT_ERROR (self, LIBRARY, SETTINGS,
        ("Could not create encoder session"), (NULL));
    return FALSE;
  }

  return TRUE;
}

static gboolean
gst_ama_enc_copy_output_frame (GstAMAEnc * self, GstBuffer * outbuf,
    XmaDataBuffer * buf)
{
  GstMapInfo map_info = GST_MAP_INFO_INIT;

  if (!gst_buffer_map (outbuf, &map_info, GST_MAP_WRITE)) {
    GST_ERROR_OBJECT (self, "Failed to map output buffer!");
    return FALSE;
  }

  memcpy (map_info.data, buf->data.buffer, map_info.size);

  gst_buffer_unmap (outbuf, &map_info);

  return TRUE;
}

static GstFlowReturn
gst_ama_enc_receive_xma_frame (GstAMAEnc * self)
{
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  GstFlowReturn result = GST_FLOW_ERROR;
  GstVideoCodecFrame *frame = NULL;
  XmaDataBuffer *buf;
  int32_t size;
  int32_t xmares;

  buf = xma_data_buffer_alloc (priv->xma->xma, 0, TRUE);

  xmares = xma_enc_session_recv_data (priv->encoder_session, buf, &size);

  if (xmares == XMA_RESEND_AND_RECV) {
    result = GST_FLOW_OK;
    goto out;
  } else if (xmares == XMA_EOS) {
    result = GST_FLOW_EOS;
    goto out;
  }

  if (xmares != XMA_SUCCESS) {
    GST_ELEMENT_ERROR (self, STREAM, ENCODE,
        ("Could not receive encoded frame: %d", xmares), (NULL));
    goto out;
  }

  frame = gst_video_encoder_get_oldest_frame (GST_VIDEO_ENCODER (self));
  if (!frame) {
    GST_WARNING_OBJECT (self, "Received XMA data but no encoder frame in queue.");
    result = GST_FLOW_OK;
    goto out;
  }

  result = gst_video_encoder_allocate_output_frame (GST_VIDEO_ENCODER (self),
      frame, size);
  if (result != GST_FLOW_OK) {
    GST_ELEMENT_ERROR (self, STREAM, ENCODE,
        ("Could not allocate output frame"), (NULL));
    goto out;
  }

  if (!gst_ama_enc_copy_output_frame (self, frame->output_buffer, buf)) {
    GST_ELEMENT_ERROR (self, STREAM, ENCODE,
        ("Could not allocate output frame"), (NULL));
    goto out;
  }

  result = gst_video_encoder_finish_frame (GST_VIDEO_ENCODER (self),
      g_steal_pointer (&frame));

out:
  xma_data_buffer_free (buf);
  if (frame) {
    gst_video_codec_frame_unref (frame);
  }

  return result;
}

static void
gst_ama_enc_add_dynamic_encode_parameters (GstAMAEnc * self, GstStructure * s,
    XmaFrame ** frame)
{
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  XmaFrameSideData *sidedata;
  XmaDynamicEncParams_v2 *params;
  XmaFrame *temp_frame;
  guint uint_val;
  gboolean bool_val;

  GST_LOG_OBJECT (self, "Using dynamic encode parameters %"GST_PTR_FORMAT, s);

  sidedata = xma_side_data_alloc (priv->xma->xma,
      XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS, XMA_HOST_BUFFER_TYPE,
      sizeof (XmaDynamicEncParams_v2));

  params = (XmaDynamicEncParams_v2*)sidedata->host;

  if (gst_structure_get_uint (s, "spatial-aq-gain", &uint_val)) {
    params->v1.spatial_aq_gain = uint_val;
    params->v1.is_spatial_aq_gain_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "temporal-aq-gain", &uint_val)) {
    params->v1.temporal_aq_gain = uint_val;
    params->v1.is_temporal_aq_gain_changed = TRUE;
  }
  if (gst_structure_get_boolean (s, "spatial-aq-mode", &bool_val)) {
    params->v1.spatial_aq_mode = bool_val;
    params->v1.is_spatial_mode_changed = TRUE;
  }
  if (gst_structure_get_boolean (s, "temporal-aq-mode", &bool_val)) {
    params->v1.temporal_aq_mode = bool_val;
    params->v1.is_temporal_mode_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "bitrate-kbps", &uint_val)) {
    params->v1.bit_rate_kbps = uint_val;
    params->v1.is_bit_rate_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "num-b-frames", &uint_val)) {
    params->v1.num_b_frames = uint_val;
    params->v1.is_b_frames_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "min-qp", &uint_val)) {
    params->v1.min_qp = uint_val;
    params->v1.is_min_qp_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "max-qp", &uint_val)) {
    params->v1.max_qp = uint_val;
    params->v1.is_max_qp_changed = TRUE;
  }

  if (gst_structure_get_uint (s, "min-bitrate-kbps", &uint_val)) {
    params->min_bit_rate_kbps = uint_val;
    params->is_min_bit_rate_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "max-bitrate-kbps", &uint_val)) {
    params->max_bit_rate_kbps = uint_val;
    params->is_max_bit_rate_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "qp", &uint_val)) {
    params->qp = uint_val;
    params->is_qp_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "qp-i-offset", &uint_val)) {
    params->qp_i_offset = uint_val;
    params->is_qp_i_offset_changed = TRUE;
  }
  if (gst_structure_get_uint (s, "qp-b-offset", &uint_val)) {
    params->qp_b_offset = uint_val;
    params->is_qp_b_offset_changed = TRUE;
  }

  temp_frame = *frame;
  *frame = xma_frame_clone(priv->xma->xma, *frame);
  xma_frame_dec_ref(temp_frame);
  xma_frame_add_side_data (*frame, sidedata);
  xma_side_data_dec_ref (sidedata);
}

static gboolean
gst_ama_enc_read_encoding_params_file (GstAMAEnc * self,
    GstVideoCodecFrame * frame, GstStructure * enc_params)
{
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  if (priv->encoding_params_stream) {
    XmaDynamicEncParams_v2 params;
    int ret;

    ret = ama_encoder_get_dyn_params_v2(gst_xma_get_log_handle (),
        priv->encoding_params_stream, frame->system_frame_number, &params);
    if (ret == 1) {
      if (params.v1.is_spatial_aq_gain_changed) {
        gst_structure_set (enc_params, "spatial-aq-gain", G_TYPE_UINT,
            params.v1.spatial_aq_gain, NULL);
      }
      if (params.v1.is_temporal_aq_gain_changed) {
        gst_structure_set (enc_params, "temporal-aq-gain", G_TYPE_UINT,
            params.v1.temporal_aq_gain, NULL);
      }
      if (params.v1.is_spatial_mode_changed) {
        gst_structure_set (enc_params, "spatial-aq-mode", G_TYPE_BOOLEAN,
            params.v1.spatial_aq_mode, NULL);
      }
      if (params.v1.is_temporal_mode_changed) {
        gst_structure_set (enc_params, "temporal-aq-mode", G_TYPE_BOOLEAN,
            params.v1.temporal_aq_mode, NULL);
      }
      if (params.v1.is_bit_rate_changed) {
        gst_structure_set (enc_params, "bitrate-kbps", G_TYPE_UINT,
            params.v1.bit_rate_kbps, NULL);
      }
      if (params.v1.is_b_frames_changed) {
        gst_structure_set (enc_params, "num-b-frames", G_TYPE_UINT,
            params.v1.num_b_frames, NULL);
      }
      if (params.v1.is_min_qp_changed) {
        gst_structure_set (enc_params, "min-qp", G_TYPE_UINT,
            params.v1.min_qp, NULL);
      }
      if (params.v1.is_max_qp_changed) {
        gst_structure_set (enc_params, "max-qp", G_TYPE_UINT,
            params.v1.max_qp, NULL);
      }
      if (params.is_qp_changed) {
        gst_structure_set (enc_params, "qp", G_TYPE_UINT,
            params.qp, NULL);
      }
      if (params.is_qp_i_offset_changed) {
        gst_structure_set (enc_params, "qp-i-offset", G_TYPE_UINT,
            params.qp_i_offset, NULL);
      }
      if (params.is_qp_b_offset_changed) {
        gst_structure_set (enc_params, "qp-b-offset", G_TYPE_UINT,
            params.qp_b_offset, NULL);
      }
      if (params.is_min_bit_rate_changed) {
        gst_structure_set (enc_params, "min-bitrate-kbps", G_TYPE_UINT,
            params.min_bit_rate_kbps, NULL);
      }
      if (params.is_max_bit_rate_changed) {
        gst_structure_set (enc_params, "max-bitrate-kbps", G_TYPE_UINT,
            params.max_bit_rate_kbps, NULL);
      }
    }
  }

  return TRUE;
}

static GstFlowReturn
gst_ama_enc_handle_frame (GstVideoEncoder * encoder,
    GstVideoCodecFrame * frame)
{
  GstAMAEnc *self = GST_AMA_ENC (encoder);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  GstMemory *mem;
  GstXMAFrame *xmaframe;
  XmaFrame *xframe;
  int32_t xmares;
  GstStructure *dyn_enc_params;
  GstBuffer *dummyinput;

  mem = gst_buffer_peek_memory (frame->input_buffer, 0);
  if (!gst_is_xma_frame (mem)) {
    GST_ELEMENT_ERROR (self, STREAM, ENCODE,
        ("invalid input video buffer"), (NULL));
    goto error;
  }

  dyn_enc_params = gst_structure_new_empty ("ama-dyn-enc-params");

  if (!gst_ama_enc_read_encoding_params_file (self, frame, dyn_enc_params)) {
    GST_ELEMENT_ERROR (self, RESOURCE, READ, (NULL),
        ("Could not read dynamic encoding parameters file"));
    gst_clear_structure (&dyn_enc_params);
    goto error;
  }

  g_signal_emit (self, signals[FRAME_RECEIVED_SIGNAL], 0, frame,
      dyn_enc_params);

  xmaframe = GST_XMA_FRAME_CAST (mem);

  if (GST_VIDEO_CODEC_FRAME_IS_FORCE_KEYFRAME (frame)) {
    GST_INFO_OBJECT (encoder, "Forcing key frame");
    xmaframe->frame->is_idr = TRUE;
  }

  xframe = xmaframe->frame;
  xma_frame_inc_ref (xframe);
  if (gst_structure_n_fields (dyn_enc_params) > 0) {
    gst_ama_enc_add_dynamic_encode_parameters (self, dyn_enc_params,
        &xframe);
  }

  gst_clear_structure (&dyn_enc_params);

  xmares = xma_enc_session_send_frame (priv->encoder_session, xframe);

  if (xmares != XMA_SUCCESS && xmares != XMA_SEND_MORE_DATA) {
    GST_ELEMENT_ERROR (self, STREAM, ENCODE, (NULL), ("failed to encode frame"));
    goto error;
  }

  /* Copy buffer metadata, which will be needed later in finish_frame function,
   * and unref the original input buffer in order to shorten the lifetime of
   * the XmaFrame it wraps, allowing its memory to be reused asap. */
  dummyinput = gst_buffer_new ();
  gst_buffer_copy_into (dummyinput, frame->input_buffer,
      GST_BUFFER_COPY_METADATA, 0, -1);
  g_clear_pointer (&frame->input_buffer, gst_buffer_unref);
  frame->input_buffer = dummyinput;

  gst_video_codec_frame_unref (frame);

  if (xmares == XMA_SEND_MORE_DATA) {
    return GST_FLOW_OK;
  }

  return gst_ama_enc_receive_xma_frame (self);

error:
  return GST_FLOW_ERROR;
}

static GstFlowReturn
gst_ama_enc_finish (GstVideoEncoder * encoder)
{
  GstAMAEnc *self = GST_AMA_ENC (encoder);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  GstFlowReturn res;

  if (!priv->encoder_session) {
    return GST_FLOW_OK;
  }

  if (xma_enc_session_send_frame (priv->encoder_session, NULL) != XMA_SUCCESS) {
    GST_ELEMENT_ERROR (self, CORE, FAILED,
        ("could not send EOF to encoder session"), (NULL));
    return GST_FLOW_ERROR;
  }

  do {
    res = gst_ama_enc_receive_xma_frame (self);
  } while (res == GST_FLOW_OK);

  return res;
}

static void
gst_ama_enc_init (GstAMAEnc * enc)
{
}

static void
gst_ama_enc_set_property (GObject * object, guint prop_id, const GValue * value,
    GParamSpec * pspec)
{
  GstAMAEnc *self = GST_AMA_ENC (object);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  switch (prop_id) {
    case PROP_PRESET:
      priv->preset = g_value_get_enum (value);
      break;
    case PROP_BITRATE:
      priv->bitrate = g_value_get_uint (value);
      break;
    case PROP_MIN_QP:
      priv->min_qp = g_value_get_int (value);
      break;
    case PROP_MAX_QP:
      priv->max_qp = g_value_get_int (value);
      break;
    case PROP_GOP_LENGTH:
      priv->gop_length = g_value_get_int (value);
      break;
    case PROP_IDR_INTERVAL:
      priv->idr_interval = g_value_get_uint (value);
      break;
    case PROP_LOOKAHEAD_DEPTH:
      priv->lookahead_depth = g_value_get_int (value);
      break;
    case PROP_SLICE:
      priv->slice = g_value_get_int (value);
      break;
    case PROP_RATE_CONTROL:
      priv->rate_control = g_value_get_enum (value);
      break;
    case PROP_MAX_BITRATE:
      priv->max_bitrate = g_value_get_int64 (value);
      break;
    case PROP_B_FRAMES:
      priv->b_frames = g_value_get_int (value);
      break;
    case PROP_FORCED_IDR:
      priv->forced_idr = g_value_get_boolean (value) ? 1 : 0;
      break;
    case PROP_QP:
      priv->qp = g_value_get_int (value);
      break;
    case PROP_QP_MODE:
      priv->qp_mode = g_value_get_enum (value);
      break;
    case PROP_SPATIAL_AQ:
      priv->spatial_aq = g_value_get_int (value);
      break;
    case PROP_SPATIAL_AQ_GAIN:
      priv->spatial_aq_gain = g_value_get_uint (value);
      break;
    case PROP_TEMPORAL_AQ:
      priv->temporal_aq = g_value_get_int (value);
      break;
    case PROP_TEMPORAL_AQ_GAIN:
      priv->temporal_aq_gain = g_value_get_uint (value);
      break;
    case PROP_TUNE_METRICS:
      priv->tune_metrics = g_value_get_enum (value);
      break;
    case PROP_TIER:
      priv->tier = g_value_get_enum (value);
      break;
    case PROP_LATENCY_LOGGING:
      priv->latency_logging = g_value_get_boolean (value) ? 1 : 0;
      break;
    case PROP_CRF:
      priv->crf = g_value_get_boolean (value) ? 1 : 0;
      break;
    case PROP_DYNAMIC_GOP:
      priv->dynamic_gop = g_value_get_int (value);
      break;
    case PROP_ENCODING_PARAMS_FILE:
      g_clear_pointer (&priv->encoding_params_file, g_free);
      priv->encoding_params_file = g_value_dup_string (value);
      break;
    case PROP_LATENCY_MS:
      priv->latency_ms = g_value_get_int (value);
      break;
    case PROP_BUFSIZE:
      priv->bufsize = g_value_get_int (value);
      break;
    case PROP_NO_LOWLAT_BFRAMES:
      priv->no_low_latency_b_frames = g_value_get_int (value);
      break;
    case PROP_MIN_BITRATE:
      priv->min_bitrate = g_value_get_int64 (value);
      break;
    case PROP_QP_I_OFFSET:
      priv->qp_i_offset = g_value_get_int (value);
      break;
    case PROP_QP_B_OFFSET:
      priv->qp_b_offset = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_enc_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstAMAEnc *self = GST_AMA_ENC (object);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  switch (prop_id) {
    case PROP_DEVICE:
      if (priv->xma) {
        g_value_set_int (value, priv->xma->device);
      } else {
        g_value_set_int (value, GST_XMA_DEVICE_UNSPECIFIED);
      }
      break;
    case PROP_PRESET:
      g_value_set_enum (value, priv->preset);
      break;
    case PROP_BITRATE:
      g_value_set_uint (value, priv->bitrate);
      break;
    case PROP_MIN_QP:
      g_value_set_int (value, priv->min_qp);
      break;
    case PROP_MAX_QP:
      g_value_set_int (value, priv->max_qp);
      break;
    case PROP_GOP_LENGTH:
      g_value_set_int (value, priv->gop_length);
      break;
    case PROP_IDR_INTERVAL:
      g_value_set_uint (value, priv->idr_interval);
      break;
    case PROP_LOOKAHEAD_DEPTH:
      g_value_set_int (value, priv->lookahead_depth);
      break;
    case PROP_SLICE:
      g_value_set_int (value, priv->slice);
      break;
    case PROP_RATE_CONTROL:
      g_value_set_enum (value, priv->rate_control);
      break;
    case PROP_MAX_BITRATE:
      g_value_set_int64 (value, priv->max_bitrate);
      break;
    case PROP_B_FRAMES:
      g_value_set_int (value, priv->b_frames);
      break;
    case PROP_FORCED_IDR:
      g_value_set_boolean (value, priv->forced_idr);
      break;
    case PROP_QP:
      g_value_set_int (value, priv->qp);
      break;
    case PROP_QP_MODE:
      g_value_set_enum (value, priv->qp_mode);
      break;
    case PROP_SPATIAL_AQ:
      g_value_set_int(value, priv->spatial_aq);
      break;
    case PROP_SPATIAL_AQ_GAIN:
      g_value_set_uint (value, priv->spatial_aq_gain);
      break;
    case PROP_TEMPORAL_AQ:
      g_value_set_int (value, priv->temporal_aq);
      break;
    case PROP_TEMPORAL_AQ_GAIN:
      g_value_set_uint (value, priv->temporal_aq_gain);
      break;
    case PROP_TUNE_METRICS:
      g_value_set_enum (value, priv->tune_metrics);
      break;
    case PROP_TIER:
      g_value_set_enum (value, priv->tier);
      break;
    case PROP_LATENCY_LOGGING:
      g_value_set_boolean (value, priv->latency_logging);
      break;
    case PROP_CRF:
      g_value_set_boolean (value, priv->crf);
      break;
    case PROP_DYNAMIC_GOP:
      g_value_set_int (value, priv->dynamic_gop);
      break;
    case PROP_ENCODING_PARAMS_FILE:
      g_value_set_string (value, priv->encoding_params_file);
      break;
    case PROP_LATENCY_MS:
      g_value_set_int (value, priv->latency_ms);
      break;
    case PROP_BUFSIZE:
      g_value_set_int (value, priv->bufsize);
      break;
    case PROP_NO_LOWLAT_BFRAMES:
      g_value_set_int (value, priv->no_low_latency_b_frames);
      break;
    case PROP_MIN_BITRATE:
      g_value_set_int64 (value, priv->min_bitrate);
      break;
    case PROP_QP_I_OFFSET:
      g_value_set_int (value, priv->qp_i_offset);
      break;
    case PROP_QP_B_OFFSET:
      g_value_set_int (value, priv->qp_b_offset);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_enc_finalize (GObject * object)
{
  GstAMAEnc *self = GST_AMA_ENC (object);
  GstAMAEncPrivate *priv = gst_ama_enc_get_instance_private (self);

  g_clear_pointer (&priv->encoding_params_file, g_free);

  G_OBJECT_CLASS (gst_ama_enc_parent_class)->finalize (object);
}

static void
gst_ama_enc_class_init (GstAMAEncClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
  GstVideoEncoderClass *encoder_class = GST_VIDEO_ENCODER_CLASS (klass);

  gobject_class->set_property = gst_ama_enc_set_property;
  gobject_class->get_property = gst_ama_enc_get_property;
  gobject_class->finalize = gst_ama_enc_finalize;

  encoder_class->start = gst_ama_enc_start;
  encoder_class->stop = gst_ama_enc_stop;
  encoder_class->set_format = gst_ama_enc_set_format;
  encoder_class->handle_frame = gst_ama_enc_handle_frame;
  encoder_class->finish = gst_ama_enc_finish;

  g_object_class_install_property (gobject_class, PROP_DEVICE,
      g_param_spec_int ("device", "XMA device",
          "Numeric ID of an XMA device (-1 = unspecified)",
          -1, G_MAXINT32, GST_XMA_DEVICE_UNSPECIFIED,
          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_PRESET,
      g_param_spec_enum ("preset", "Preset",
          "Preset (can be medium, slow or fast)", GST_TYPE_AMA_ENC_PRESET,
          GST_AMA_ENC_PRESET_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_BITRATE,
      g_param_spec_uint ("bitrate", "Bitrate",
          "Fixed output bitrate in bits per second (0 = disabled)",
          0, G_MAXUINT32, GST_AMA_ENC_BITRATE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_GOP_LENGTH,
      g_param_spec_int ("gop-length", "GOP length",
          "Maximum distance between two consecutive I frames (-1 = auto)",
          -1, G_MAXINT32, GST_AMA_ENC_GOP_LENGTH_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_IDR_INTERVAL,
      g_param_spec_uint ("idr-interval", "IDR frame interval",
          "The periodicity of IDR frames",
          0, G_MAXUINT32, GST_AMA_ENC_IDR_INTERVAL_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_LOOKAHEAD_DEPTH,
      g_param_spec_int ("lookahead-depth", "Lookahead depth",
          "Number of frames in the future of the currently encoded frame that "
          "the encoder will analyze and take into account (-1 = auto-select "
          "based on fps)",
          -1, GST_AMA_ENC_LOOKAHEAD_DEPTH_MAX,
          GST_AMA_ENC_LOOKAHEAD_DEPTH_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_SLICE,
      g_param_spec_int ("slice", "Device slice number",
          "Which slice of the XMA device to use (-1 = auto)",
          -1, GST_AMA_NUM_SLICES - 1,
          GST_AMA_ENC_SLICE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_RATE_CONTROL,
      g_param_spec_enum ("rate-control", "Rate control mode",
          "Rate control mode to use for encoding", GST_TYPE_AMA_ENC_RATE_CONTROL,
          GST_AMA_ENC_RC_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_MAX_BITRATE,
      g_param_spec_int64 ("max-bitrate", "Maximum bitrate",
          "Use to limit encoding bitrate if you have not specified \"bitrate\" "
          "parameter",
          -1, 35000000000,
          GST_AMA_ENC_MAX_BITRATE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_FORCED_IDR,
      g_param_spec_boolean ("forced-idr", "Forced IDR",
          "Encode all intra frames as IDR",
          GST_AMA_ENC_FORCED_IDR_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_QP_MODE,
      g_param_spec_enum ("qp-mode", "QP control mode",
          "QP control mode", GST_TYPE_AMA_ENC_QP_MODE,
          GST_AMA_ENC_QP_MODE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_SPATIAL_AQ,
      g_param_spec_int ("spatial-aq", "Spatial AQ",
          "Enable spatial AQ (-1 = auto)",
          -1,1,
          GST_AMA_ENC_SPATIAL_AQ_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_SPATIAL_AQ_GAIN,
      g_param_spec_uint ("spatial-aq-gain", "Spatial AQ gain",
          "Spatial AQ gain scale factor",
          0, XMA_ENC_MAX_SPATIAL_AQ_GAIN,
          GST_AMA_ENC_SPATIAL_AQ_GAIN_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_TEMPORAL_AQ,
      g_param_spec_int ("temporal-aq", "Temporal AQ",
          "Enable temporal AQ (-1 = auto)",
          -1,1,
          GST_AMA_ENC_TEMPORAL_AQ_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_TEMPORAL_AQ_GAIN,
      g_param_spec_uint ("temporal-aq-gain", "Temporal AQ gain",
          "Temporal AQ gain scale factor",
          0, XMA_ENC_MAX_TEMPORAL_AQ_GAIN,
          GST_AMA_ENC_TEMPORAL_AQ_GAIN_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_TUNE_METRICS,
      g_param_spec_enum ("tune-metrics", "Tune metrics",
          "Tune VQ for objective metrics",
          GST_TYPE_AMA_ENC_TUNE_METRICS,
          GST_AMA_ENC_TUNE_METRICS_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_TIER,
      g_param_spec_enum ("tier", "Tier",
          "Tier",
          GST_TYPE_AMA_ENC_TIER,
          GST_AMA_ENC_TIER_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_LATENCY_LOGGING,
      g_param_spec_boolean ("latency-logging", "Latency logging",
          "Log latency info to syslog",
          GST_AMA_ENC_LATENCY_LOGGING_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_CRF,
      g_param_spec_boolean ("crf", "Constant Rate Factor",
          "Enable CRF mode",
          GST_AMA_ENC_CRF_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_DYNAMIC_GOP,
      g_param_spec_int ("dynamic-gop", "Dynamic GOP",
          "Enable dynamically determining the optimal number of B frames, based on the content. (-1 = auto)",
          -1,1,
          GST_AMA_ENC_DYNAMIC_GOP_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_ENCODING_PARAMS_FILE,
      g_param_spec_string ("encoding-params-file", "Encoding parameters file",
          "If specified, encoder will read dynamic parameters from this file",
          GST_AMA_ENCODING_PARAMS_FILE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_LATENCY_MS,
      g_param_spec_int ("latency-ms", "Encoder latency",
          "Encoder latency in milliseconds (-1 = auto)",
          XMA_ENC_LATENCY_MS_MIN, XMA_ENC_LATENCY_MS_MAX,
          GST_AMA_ENC_LATENCY_MS_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_BUFSIZE,
      g_param_spec_int ("bufsize", "VBV Buffer Size",
          "Size of VBV buffer (in bits) (-1 = auto)",
          XMA_ENC_BUFSIZE_MIN, XMA_ENC_BUFSIZE_MAX,
          GST_AMA_ENC_BUFSIZE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_NO_LOWLAT_BFRAMES,
      g_param_spec_int ("no-bll", "No low latency Bframes",
          "No low latency Bframes (Default = -1)",
          XMA_ENC_NO_LOWLAT_BFRAMES_DEFAULT, XMA_ENC_NO_LOWLAT_BFRAMES_ENABLE,
          GST_AMA_ENC_NO_LOWLAT_BFRAMES_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

    g_object_class_install_property (gobject_class, PROP_MIN_BITRATE,
      g_param_spec_int64 ("min-bitrate", "Minimum bitrate",
          "Use to limit encoding bitrate if you have not specified \"bitrate\" "
          "parameter",
          -1, 35000000000,
          GST_AMA_ENC_MIN_BITRATE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  signals[FRAME_RECEIVED_SIGNAL] =
      g_signal_new ("frame-received", G_TYPE_FROM_CLASS (klass),
      G_SIGNAL_RUN_LAST, 0, NULL, NULL, NULL,
          G_TYPE_NONE, 2, GST_TYPE_VIDEO_CODEC_FRAME, G_TYPE_POINTER);

  gst_element_class_add_static_pad_template (element_class, &sink_template);

  GST_DEBUG_CATEGORY_INIT (gst_ama_enc_debug, "ama_enc", 0,
      "Xilinx AMA H264/H265/AV1 encoder");
}

void
gst_ama_enc_class_init_ranged_properties (GstAMAEncClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_object_class_install_property (gobject_class, PROP_B_FRAMES,
      g_param_spec_int ("b-frames", "Number of B-frames",
          "Number of B-frames between two consecutive P-frames",
          XMA_ENC_MIN_BF_VALUE, klass->b_frames_max, GST_AMA_ENC_B_FRAMES_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_QP,
      g_param_spec_int ("qp", "Fixed Quantization value",
          "Fixed quantization value (-1 to disable)",
          -1, klass->qp_max, GST_AMA_ENC_QP_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_MIN_QP,
      g_param_spec_int ("min-qp", "min Quantization value",
          "Minimum QP value allowed for the rate control (-1 = auto)",
          XMA_ENC_QP_MIN_VALUE, klass->min_qp_max, GST_AMA_ENC_MIN_QP_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_MAX_QP,
      g_param_spec_int ("max-qp", "max Quantization value",
          "Maximum QP value allowed for the rate control (-1 = auto)",
          XMA_ENC_QP_MIN_VALUE, klass->max_qp_max, klass->max_qp_default,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_QP_I_OFFSET,
      g_param_spec_int ("qp-i-offset", "Quantization offset value - I",
          "Quantization offset value - I (-1 = auto)",
          XMA_ENC_QP_I_OFFSET_MIN_VALUE, XMA_ENC_QP_I_OFFSET_MAX_VALUE, XMA_ENC_PARAM_QP_I_OFFSET_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_QP_B_OFFSET,
      g_param_spec_int ("qp-b-offset", "Quantization offset value - P/B",
          "Quantization offset value - P/B (-1 = auto)",
          XMA_ENC_QP_B_OFFSET_MIN_VALUE, XMA_ENC_QP_B_OFFSET_MAX_VALUE, XMA_ENC_PARAM_QP_B_OFFSET_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));
}
